Castbars = LibStub("AceAddon-3.0"):NewAddon("Castbars");

Castbars.SharedMedia = LibStub("LibSharedMedia-3.0");

function Castbars:FrameTextureRestore(frame)
    local barTexture = self.SharedMedia:Fetch('statusbar', self.db.profile[frame.configName]["Texture"]);
    local borderTexture = self.SharedMedia:Fetch('border', self.db.profile[frame.configName]["Border"]);
    if (barTexture) then
        frame.statusBar:SetStatusBarTexture(barTexture);
    end
    if (borderTexture) then
        frame.backdrop:SetBackdrop({edgeFile = borderTexture, edgeSize = 16, tileSize = 16, insets = {bottom = 5, top = 5, right = 5, left = 5}});
        local brightness = self.db.profile[frame.configName]["BorderBrightness"];
        frame.backdrop:SetBackdropBorderColor(brightness, brightness, brightness, 1.0);
    end
end

function Castbars:FrameTextureRestoreAll()
    for i, frame in pairs(self.frames) do
        self:FrameTextureRestore(frame)
    end
end

function Castbars:FrameLayoutRestore(frame)
    -- Position
    local position = self.db.profile[frame.configName]["Position"];
    if (frame.dragable and position) then
        frame:ClearAllPoints();
        frame:SetPoint(position.point, position.parent, position.relpoint, position.x, position.y);
    end

    -- Texture
    self:FrameTextureRestore(frame)

    -- Dimensions
    frame:SetWidth(self.db.profile[frame.configName]["Width"]);
    frame:SetHeight(self.db.profile[frame.configName]["Height"]);

    -- Icon visibility and position
    if (frame.icon) then
        if (self.db.profile[frame.configName]["ShowIcon"]) then
            frame.icon:ClearAllPoints();
            if (self.db.profile[frame.configName]["Border"] == "None") then
                frame.icon:SetPoint("RIGHT", frame, "LEFT", 0, 0);
                frame.icon:SetHeight(self.db.profile[frame.configName]["Height"]);
                frame.icon:SetWidth(self.db.profile[frame.configName]["Height"]);
            else
                frame.icon:SetPoint("RIGHT", frame, "LEFT", -5, 0);
                frame.icon:SetHeight(self.db.profile[frame.configName]["Height"] + 5);
                frame.icon:SetWidth(self.db.profile[frame.configName]["Height"] + 5);
            end
            frame.icon:Show();
        else
            frame.icon:Hide();
        end
    end

    -- Visibility
    if (self.db.profile[frame.configName]["Show"] == false) then
        frame.showCastbar = false;
    else
        frame.showCastbar = true;
    end
end

function Castbars:FrameLayoutRestoreAll()
    for i, frame in pairs(self.frames) do
        self:FrameLayoutRestore(frame)
    end
end

function Castbars:FrameCustomize(frame)
    local frameName = frame:GetName();

    if (frameName:sub(1, 11) == "MirrorTimer") then
        frame.type = "Mirror";
        frame.index = tonumber(frameName:sub(12,12));
        frame.statusBar = _G[frameName .. "StatusBar"];
        frame.configName = "MirrorTimer";
        if (frame.index == 1) then
            frame.dragable = true;
        end
    elseif (frameName == "CastingBarFrame" or frameName == "TargetCastingBarFrame") then
        frame.type = "Castbar";
        frame.statusBar = frame;
        frame.configName = frameName;
        frame.dragable = true;
        frame.icon = _G[frameName .. "Icon"];
    end

    -- Make dragable
    if (frame.dragable) then
        frame:EnableMouse(true);
        frame:SetMovable(true);
        frame:RegisterForDrag("LeftButton");
        frame:SetScript("OnDragStart", function(frame)
            if (self.ConfigMode) then
                frame:StartMoving();
            end
        end);
        frame:SetScript("OnDragStop", function(frame)
            frame:StopMovingOrSizing();
            if (not self.db.profile[frame.configName]["Position"]) then
                self.db.profile[frame.configName]["Position"] = {};
            end
            local position = self.db.profile[frame.configName]["Position"];
            position.point, position.parent, position.relpoint, position.x, position.y = frame:GetPoint();
        end);
    end

    -- Adjust spark position
    frame.spark = _G[frameName .. "Spark"];
    if (frame.spark) then
        local setPoint = frame.spark.SetPoint;
        frame.spark.SetPoint = function(self, point, relativeFrame, realativePoint, x, y)
            setPoint(self, point, relativeFrame, realativePoint, x, 0);
        end
    end

    -- Adjust text position
    frame.text = _G[frameName .. "Text"];
    frame.text:ClearAllPoints();
    frame.text:SetPoint("CENTER", frame.statusBar, "CENTER", 0, 0);

    -- Remove the border
    frame.border = _G[frameName .. "Border"];
    frame.border:SetTexture();

    -- Remove the flash
    frame.flash = _G[frameName .. "Flash"];
    if (frame.flash) then
        frame.flash:SetTexture();
    end

    -- Create backdrop
    frame.backdrop = CreateFrame("Frame", frameName .. "Backdrop", frame);
    if (frame.type == "Castbar") then
        frame.backdrop:SetPoint("CENTER", frame, "CENTER", 0, 0);
    elseif (frame.type == "Mirror") then
        frame.backdrop:SetPoint("CENTER", frame, "CENTER", 0, 3);
    end

    if (frame.type == "Castbar") then
        -- Add timer text
        frame.timer = frame:CreateFontString(nil, "ARTWORK", "GameFontHighlight");
        frame.timer:SetPoint("RIGHT", frame, "RIGHT", -5, 0);
        frame.nextupdate = 0.1;

        -- Add latency indication texture
        frame.latency = frame:CreateTexture(nil, "BACKGROUND");
        frame.latency:SetHeight(frame:GetHeight());
        frame.latency:SetPoint("RIGHT", frame, "RIGHT", 0, 0);
        frame.latency:SetTexture(1, 0, 0, 1);
    end

    if (frame.type == "Mirror") then
        for i, region in pairs({frame:GetRegions()}) do
            if (region.GetTexture and region:GetTexture() == "SolidTexture") then
                frame.shade = region;
            end
        end
    end

    -- Automatically adjust the height of sub elements
    local setHeight = frame.SetHeight;
    if (frame.type == "Castbar") then
        frame.SetHeight = function(self, height)
            frame.backdrop:SetHeight(height + 10);
            frame.spark:SetHeight(2.5 * height);
            frame.latency:SetHeight(height);
            setHeight(self, height);
        end
    elseif (frame.type == "Mirror") then
        frame.SetHeight = function(self, height)
            frame.backdrop:SetHeight(height + 10);
            frame.statusBar:SetHeight(height);
            frame.shade:SetHeight(height);
            setHeight(self, height + 10);
        end
    end

    -- Automatically adjust the width of sub elements
    local setWidth = frame.SetWidth
    if (frame.type == "Castbar") then
        frame.SetWidth = function(self, width)
            frame.backdrop:SetWidth(width + 10);
            setWidth(self, width);
        end
    elseif (frame.type == "Mirror") then
        frame.SetWidth = function(self, width)
            frame.backdrop:SetWidth(width + 10);
            frame.statusBar:SetWidth(width);
            frame.shade:SetWidth(width);
            setWidth(self, width + 10);
        end
    end

    frame:Hide();
end

function Castbars:FrameCustomizeAll()
    for i, frame in pairs(self.frames) do
        self:FrameCustomize(frame)
    end
end

function Castbars:GetOptionsTableForBar(frameConfigName, friendlyName, order, enableDisable, showHideIcon)
    local optionstable =
    {
        type = "group",
        name = friendlyName .. " Options",
        guiInline = true,
        order = order,
        args =
        {
            width =
            {
                name = "Width",
                type = "range",
                order = 3,
                desc = "Set the width of the " .. friendlyName,
                min = 100,
                max = 300,
                get = function() return self.db.profile[frameConfigName]["Width"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName]["Width"] = value;
                    self:FrameLayoutRestoreAll();
                end,
            },
            height =
            {
                name = "Height",
                type = "range",
                order = 4,
                desc = "Set the height of the " .. friendlyName,
                min = 10,
                max = 50,
                get = function() return self.db.profile[frameConfigName]["Height"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName]["Height"] = value;
                    self:FrameLayoutRestoreAll();
                end,
            },
            texture =
            {
                name = "Texture",
                type = 'select',
                order = 5,
                desc = "Select texture to use for the " .. friendlyName,
                dialogControl = 'LSM30_Statusbar',
                values = AceGUIWidgetLSMlists.statusbar,
                get = function() return self.db.profile[frameConfigName]["Texture"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName]["Texture"] = value;
                    self:FrameLayoutRestoreAll();
                end,
            },
            border =
            {
                name = "Border",
                type = 'select',
                order = 6,
                desc = "Select border to use for the " .. friendlyName,
                dialogControl = 'LSM30_Border',
                values = AceGUIWidgetLSMlists.border,
                get = function() return self.db.profile[frameConfigName]["Border"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName]["Border"] = value;
                    self:FrameLayoutRestoreAll();
                end,
            },
            borderbrightness =
            {
                name = "Border Brightness",
                type = "range",
                order = 7,
                desc = "Set the brightness of the border of the " .. friendlyName,
                min = 0.0,
                max = 1.0,
                isPercent = true,
                get = function() return self.db.profile[frameConfigName]["BorderBrightness"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName]["BorderBrightness"] = value;
                    self:FrameLayoutRestoreAll();
                end,
            },
        },
    };
    if (enableDisable) then
        optionstable.args.enable =
        {
            name = "Enable",
            type = "toggle",
            order = 1,
            desc = "Toggles display of the " .. friendlyName,
            get = function() return self.db.profile[frameConfigName]["Show"] end,
            set = function()
                self.db.profile[frameConfigName]["Show"] = not self.db.profile[frameConfigName]["Show"];
                self:FrameLayoutRestoreAll();
            end,
        }
    end
    if (showHideIcon) then
        optionstable.args.showicon =
        {
            name = "Show Icon",
            type = "toggle",
            order = 2,
            desc = "Toggles display of the icon at the left side of the bar",
            get = function() return self.db.profile[frameConfigName]["ShowIcon"] end,
            set = function()
                self.db.profile[frameConfigName]["ShowIcon"] = not self.db.profile[frameConfigName]["ShowIcon"];
                self:FrameLayoutRestoreAll();
            end,
        }
    end
    return optionstable;
end

function Castbars:GetOptionsTable()
    local options =
    {
        type = "group",
        args =
        {
            general =
            {
                type = "group",
                name = "General Settings",
                handler = self,
                childGroups = "tree",
                args =
                {
                    configmode =
                    {
                        type = "group",
                        name = "Configuration Mode",
                        guiInline = true,
                        order = 0,
                        args =
                        {
                            toggleconfigmode =
                            {
                                name = "Toggle",
                                type = "execute",
                                order = 1,
                                desc = "Toggle configuration mode to allow moving bars and setting appearance options",
                                func = "Toggle",
                            },
                        },
                    },
                    player = self:GetOptionsTableForBar("CastingBarFrame", "Player Castbar", 1, true, true),
                    target = self:GetOptionsTableForBar("TargetCastingBarFrame", "Target Castbar", 2, true, true),
                    mirror = self:GetOptionsTableForBar("MirrorTimer", "Mirror Timer Bars", 3),
                },
            },
            profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db),
        },
    };
    return options;
end

function Castbars:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("CastbarsDB",
    {
        profile =
        {
            ['**'] =
            {
                Width = 195,
                Height = 13,
                Texture = "Blizzard",
                Border = "Blizzard Dialog",
                BorderBrightness = 0.5,
                Position = nil,
            },
            CastingBarFrame =
            {
                Show = true,
                ShowIcon = true,
                Position = {point = "BOTTOM", relpoint = "BOTTOM", x = 0, y = 200},
            },
            TargetCastingBarFrame =
            {
                Show = true,
                ShowIcon = true,
                Border = "None",
                Position = {point = "CENTER", relpoint = "CENTER", x = 0, y = 200},
            },
            MirrorTimer =
            {
            },
        }
    });
    self.db.RegisterCallback(self, "OnProfileChanged", "FrameLayoutRestoreAll");
    self.db.RegisterCallback(self, "OnProfileCopied", "FrameLayoutRestoreAll");
    self.db.RegisterCallback(self, "OnProfileReset", "FrameLayoutRestoreAll");

    LibStub("AceConfig-3.0"):RegisterOptionsTable("Castbars", self:GetOptionsTable());
    self.BlizzardOptionsFrameGeneral = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Castbars", nil, nil, "general");
    self.BlizzardOptionsFrameProfile = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Castbars", "Profiles", "Castbars", "profile");

    -- Prevent the UIParent from moving the CastingBarFrame around
    UIPARENT_MANAGED_FRAME_POSITIONS["CastingBarFrame"] = nil;

    -- Create target casting bar
    CreateFrame("StatusBar", "TargetCastingBarFrame", UIParent, "CastingBarFrameTemplate");
    TargetCastingBarFrame:SetScript("OnEvent", function(frame, event, ...)
        if (event == "PLAYER_TARGET_CHANGED") then
            event = "PLAYER_ENTERING_WORLD"
        end
        CastingBarFrame_OnEvent(frame, event, ...);
    end)
    TargetCastingBarFrame:RegisterEvent("PLAYER_TARGET_CHANGED");
    CastingBarFrame_OnLoad(TargetCastingBarFrame, "target", false);

    -- Setup table with all frames
    self.frames = {CastingBarFrame, TargetCastingBarFrame, MirrorTimer1, MirrorTimer2, MirrorTimer3};

    -- Customize the bars
    self:FrameCustomizeAll();

    -- Restore Frame Layout
    self:FrameLayoutRestoreAll();

    -- Listen to LibSharedMedia-3.0 callbacks
    self.SharedMedia.RegisterCallback(self, "LibSharedMedia_Registered", "FrameTextureRestoreAll")
    self.SharedMedia.RegisterCallback(self, "LibSharedMedia_SetGlobal", "FrameTextureRestoreAll")

    -- Add timer handling code
    hooksecurefunc("CastingBarFrame_OnUpdate", function(self, elapsed)
        if (self.timer) then
            if (self.nextupdate < elapsed) then
                if (self.casting) then
                    self.timer:SetText(format("%.1f", max(self.maxValue - self.value, 0)));
                elseif (self.channeling) then
                    self.timer:SetText(format("%.1f", max(self.value, 0)));
                else
                    self.timer:SetText("");
                end;
                self.nextupdate = 0.1;
            else
                self.nextupdate = self.nextupdate - elapsed;
            end;
        end;
    end);

    -- Add latency indication handling code
    hooksecurefunc("CastingBarFrame_OnShow", function(self)
        if (self.latency) then
            local min, max = self:GetMinMaxValues();
            local down, up, lag = GetNetStats();
            local latency = (lag / 1000) / (max - min);
            if (latency < 0) then latency = 0 elseif (latency > 1) then latency = 1 end
            if (self.channeling) then
                self.latency:ClearAllPoints();
                self.latency:SetPoint("LEFT", self, "LEFT", 0, 0);
            else
                self.latency:ClearAllPoints();
                self.latency:SetPoint("RIGHT", self, "RIGHT", 0, 0);
            end;
            if (max > 4) then
                self.latency:Hide();
            else
                self.latency:SetWidth(self:GetWidth() * latency);
                self.latency:Show();
            end
        end
    end);

    self.GetOptionsTableForBar = nil;
    self.GetOptionsTable = nil;
    self.FrameCustomize = nil;
    self.OnInitialize = nil;
end

--[[ ConfigMode Support ]]--

Castbars.CastingBarFrame_OnEvent = CastingBarFrame_OnEvent;
Castbars.MirrorTimerFrame_OnEvent = MirrorTimerFrame_OnEvent;
Castbars.MirrorTimer_Show = MirrorTimer_Show;
Castbars.DoNothing = function() end;

function Castbars:Show()
    self.ConfigMode = true;
    CastingBarFrame_OnEvent = self.DoNothing;
    MirrorTimerFrame_OnEvent = self.DoNothing
    MirrorTimer_Show = self.DoNothing;

    CastingBarFrameText:SetText("Player Castbar")
    CastingBarFrame:SetStatusBarColor(0.0, 1.0, 0.0);
    CastingBarFrame:SetAlpha(1)
    CastingBarFrameSpark:Hide();
    CastingBarFrameFlash:Hide();
    CastingBarFrame.fadeOut = nil;
    CastingBarFrame:Show();

    TargetCastingBarFrameText:SetText("Target Castbar")
    TargetCastingBarFrame:SetStatusBarColor(0.0, 1.0, 0.0);
    TargetCastingBarFrame:SetAlpha(1)
    TargetCastingBarFrameSpark:Hide();
    TargetCastingBarFrameFlash:Hide();
    TargetCastingBarFrame.fadeOut = nil;
    TargetCastingBarFrame:Show();

    MirrorTimer1.paused = 1;
    MirrorTimer2.paused = 1;
    MirrorTimer3.paused = 1;
    MirrorTimer1Text:SetText("Mirror Timer 1");
    MirrorTimer2Text:SetText("Mirror Timer 2");
    MirrorTimer3Text:SetText("Mirror Timer 3");
    MirrorTimer1StatusBar:SetStatusBarColor(0.0, 1.0, 0.0);
    MirrorTimer2StatusBar:SetStatusBarColor(0.0, 1.0, 0.0);
    MirrorTimer3StatusBar:SetStatusBarColor(0.0, 1.0, 0.0);
    MirrorTimer1StatusBar:SetValue(select(2, MirrorTimer1StatusBar:GetMinMaxValues()));
    MirrorTimer2StatusBar:SetValue(select(2, MirrorTimer2StatusBar:GetMinMaxValues()));
    MirrorTimer3StatusBar:SetValue(select(2, MirrorTimer3StatusBar:GetMinMaxValues()));
    MirrorTimer1:Show();
    MirrorTimer2:Show();
    MirrorTimer3:Show();
end

function Castbars:Hide()
    MirrorTimer1:Hide();
    MirrorTimer2:Hide();
    MirrorTimer3:Hide();
    CastingBarFrame:Hide();
    TargetCastingBarFrame:Hide();
    CastingBarFrame_OnEvent = self.CastingBarFrame_OnEvent;
    MirrorTimerFrame_OnEvent = self.MirrorTimerFrame_OnEvent;
    MirrorTimer_Show = self.MirrorTimer_Show;
    self.ConfigMode = false;
end

function Castbars:Toggle()
    if (self.ConfigMode) then
        self:Hide();
    else
        self:Show();
    end
end

CONFIGMODE_CALLBACKS = CONFIGMODE_CALLBACKS or {};
CONFIGMODE_CALLBACKS["Castbars"] = function(action)
    if (action == "ON") then Castbars:Show()
    elseif (action == "OFF") then Castbars:Hide() end
end
